VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "VEval"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = True
Attribute VB_Exposed = False
Option Explicit
'=====
'VEval version 1.3
'=====
'
'Variant expression evaluator for numeric expressions.
'
'This is a VB_PredeclaredId = True class, so you do not need to create
'private instances unless you want different settings from those of the
'global predeclared instance.
'
'Depends on VFormat.cls version 1.1 or newer, but only for the LCIDs enum.
'
'
'Properties
'----------
'
'   o HexOctalAllowed [Boolean, R/W]
'
'     Default is True.
'
'
'   o LCID [Long, R/W]
'
'     Default is 0, which uses the current process LCID.
'
'
'   o ParsedTypes [VarTypeBits, R/W]
'
'     Default is [_vtbValid] (vtbInteger Or vtbLong Or vtbSingle Or vtbDouble
'     Or vtbCurrency Or vtbDecimal).
'
'     Type or types to be used for parsed numeric intermediate values.
'
'     NOTE: The default normally is not what you want though.  It typically
'           results in Integer, Long, or Currency.
'
'
'Methods
'-------
'
'   o Eval(ByRef Expression As String) As Variant
'
'     Expression is not altered.
'
'     Operators: +, -, *, /, \ (integer divide), % (Mod), ^ (Power)
'
'     Return value upon success has the Variant subtype that fits.
'
'     NOTE: Default member.
'

Private Declare Sub GetMem2 Lib "msvbvm60" (ByVal Ptr As Long, ByRef Value As Integer)

Private Declare Function StrSpn Lib "shlwapi" Alias "StrSpnW" ( _
    ByVal psz As Long, _
    ByVal pszSet As Long) As Long

Private Enum HRESULT
    S_OK = 0
    E_OUTOFMEMORY = &H8007000E
    DISP_E_TYPEMISMATCH = &H80020005
    DISP_E_OVERFLOW = &H8002000A
End Enum

Private Enum NUMPRS_FLAGS
    NUMPRS_STD = &H1FFF&
    'Flags used by both dwInFlags and dwOutFlags:
    NUMPRS_LEADING_WHITE = &H1&
    NUMPRS_TRAILING_WHITE = &H2&
    NUMPRS_LEADING_PLUS = &H4&
    NUMPRS_TRAILING_PLUS = &H8&
    NUMPRS_LEADING_MINUS = &H10&
    NUMPRS_TRAILING_MINUS = &H20&
    NUMPRS_HEX_OCT = &H40&
    NUMPRS_PARENS = &H80&
    NUMPRS_DECIMAL = &H100&
    NUMPRS_THOUSANDS = &H200&
    NUMPRS_CURRENCY = &H400&
    NUMPRS_EXPONENT = &H800&
    NUMPRS_USE_ALL = &H1000&
    'Flags used by dwOutFlags only:
    NUMPRS_NEG = &H10000
    NUMPRS_INEXACT = &H20000
End Enum

Private Enum BASE_SHIFTS
    Base10 = 0
    Base8 = 3
    Base16 = 4
End Enum

Private Type NUMPARSE
    cDig As Long
    dwInFlags As NUMPRS_FLAGS
    dwOutFlags As NUMPRS_FLAGS
    cchUsed As Long
    nBaseShift As BASE_SHIFTS
    nPwr10 As Long
End Type

Private Enum VARENUM
    VT_EMPTY = 0
    VT_NULL = 1
    VT_I2 = 2
    VT_I4 = 3
    VT_R4 = 4
    VT_R8 = 5
    VT_CY = 6
    VT_DATE = 7
    VT_BSTR = 8
    VT_DISPATCH = 9
    VT_ERROR = 10
    VT_BOOL = 11
    VT_VARIANT = 12
    VT_UNKNOWN = 13
    VT_DECIMAL = 14
    VT_I1 = 16
    VT_UI1 = 17
    VT_UI2 = 18
    VT_UI4 = 19
    VT_I8 = 20
    VT_UI8 = 21
    VT_INT = 22
    VT_UINT = 23
    VT_VOID = 24
    VT_HRESULT = 25
    VT_PTR = 26
    VT_SAFEARRAY = 27
    VT_CARRAY = 28
    VT_USERDEFINED = 29
    VT_LPSTR = 30
    VT_LPWSTR = 31
    VT_RECORD = 36
    VT_INT_PTR = 37
    VT_UINT_PTR = 38
    VT_FILETIME = 64
    VT_BLOB = 65
    VT_STREAM = 66
    VT_STORAGE = 67
    VT_STREAMED_OBJECT = 68
    VT_STORED_OBJECT = 69
    VT_BLOB_OBJECT = 70
    VT_CF = 71
    VT_CLSID = 72
    VT_VERSIONED_STREAM = 73
    VT_BSTR_BLOB = &HFFF&
    VT_VECTOR = &H1000&
    VT_ARRAY = &H2000&
    VT_BYREF = &H4000&
End Enum

Private Enum VTBITS
    VTBIT_I1 = 2 ^ VT_I1
    VTBIT_UI1 = 2 ^ VT_UI1
    VTBIT_I2 = 2 ^ VT_I2
    VTBIT_UI2 = 2 ^ VT_UI2
    VTBIT_I4 = 2 ^ VT_I4
    VTBIT_UI4 = 2 ^ VT_UI4
    VTBIT_I8 = 2 ^ VT_I8
    VTBIT_UI8 = 2 ^ VT_UI8
    VTBIT_R4 = 2 ^ VT_R4
    VTBIT_R8 = 2 ^ VT_R8
    VTBIT_CY = 2 ^ VT_CY
    VTBIT_DECIMAL = 2 ^ VT_DECIMAL
End Enum

Private Declare Function VarNumFromParseNum Lib "oleaut32" ( _
    ByRef NUMPARSE As NUMPARSE, _
    ByRef rgbDig As Byte, _
    ByVal dwVtBits As VTBITS, _
    ByRef Value As Variant) As HRESULT

Private Declare Function VarParseNumFromStr Lib "oleaut32" ( _
    ByVal strIn As Long, _
    ByVal LCID As Long, _
    ByVal dwFlags As NUMPRS_FLAGS, _
    ByRef NUMPARSE As NUMPARSE, _
    ByRef rgbDig As Byte) As HRESULT

Private Const MAX_DIGITS As Long = 29
Private Const WHITE_SPACE As String = " "
Private Const PARSE_FLAGS_DEFAULT As Long = NUMPRS_LEADING_PLUS _
                                         Or NUMPRS_LEADING_MINUS _
                                         Or NUMPRS_DECIMAL _
                                         Or NUMPRS_THOUSANDS _
                                         Or NUMPRS_EXPONENT
Private Const OPERATORS As String = "()+-*/\%^"
Private Const OPERATOR_PRECEDENCES = "992233334" 'Greater is higher precedence.
Private Const PAREN_LEFT As Byte = 0
Private Const PAREN_RIGHT As Byte = 1
Private Const PARENS_LAST As Byte = PAREN_RIGHT
Private Const PLUS As Byte = 2
Private Const MINUS As Byte = 3
Private Const STAR As Byte = 4
Private Const SLASH As Byte = 5
Private Const BACKSLASH As Byte = 6
Private Const PERCENT As Byte = 7
Private Const CIRCUMFLEX As Byte = 8

Private Const Q_INITIAL_SIZE As Long = 30
Private Const Q_GROWBY As Long = 10
Private Queue() As Variant
Private QueueLast As Long

Private Const STK_INITIAL_SIZE As Long = 15
Private Const STK_GROWBY As Long = 5
Private Stack() As Variant
Private StackTop As Long

Private rgbDig() As Byte
Private NUMPARSE As NUMPARSE
Private pWhiteSpace As Long
Private OperatorVals(255) As Byte 'Very sparse table but fast lookups.
Private Const INVALID_OPERATOR As Byte = 255 'OperatorVals() has this for all but valid entries.
Private OperatorPrecedences() As Byte
Private pNextChar As Long
Private CharsLeft As Long
Private OperandExpected As Boolean

Private mParsedTypes As Long

Public Enum VarTypeBits
    vtbInteger = VTBIT_I2
    vtbLong = VTBIT_I4
    vtbSingle = VTBIT_R4
    vtbDouble = VTBIT_R8
    vtbCurrency = VTBIT_CY
    vtbDecimal = VTBIT_DECIMAL
    vtbInt = vtbInteger Or vtbLong
    vtbReal = vtbSingle Or vtbDouble
    [_vtbValid] = vtbInteger Or vtbLong Or vtbSingle Or vtbDouble Or vtbCurrency Or vtbDecimal
End Enum
#If False Then
Dim vtbInteger, vtbLong, vtbSingle, vtbDouble, vtbCurrency, vtbDecimal
Dim vtbInt, vtbReal
#End If

Public LCID As LCIDs

Public Property Get HexOctalAllowed() As Boolean
    HexOctalAllowed = CBool(NUMPARSE.dwInFlags And NUMPRS_HEX_OCT)
End Property

Public Property Let HexOctalAllowed(ByVal RHS As Boolean)
    If RHS Then
        NUMPARSE.dwInFlags = NUMPARSE.dwInFlags Or NUMPRS_HEX_OCT
    Else
        NUMPARSE.dwInFlags = NUMPARSE.dwInFlags And Not NUMPRS_HEX_OCT
    End If
End Property

Public Property Get ParsedTypes() As VarTypeBits
    ParsedTypes = mParsedTypes
End Property

Public Property Let ParsedTypes(ByVal RHS As VarTypeBits)
    mParsedTypes = RHS And [_vtbValid]
End Property

'Default member:
Public Function Eval(ByRef Expression As String) As Variant
Attribute Eval.VB_UserMemId = 0
    pNextChar = StrPtr(Expression)
    CharsLeft = Len(Expression)
    OperandExpected = True
    ParseToPostfix
    Eval = EvaluateOutput()
End Function

Private Function EvaluateOutput() As Variant
    Dim Q As Long
    Dim Token As Variant

    'Queue has been populated and stack has been emptied.
    For Q = 0 To QueueLast
        Token = Queue(Q)
        If VarType(Token) = vbByte Then
            If StackTop < 1 Then
                If Token = MINUS Then
                    'Unary minus:
                    Token = -Stack(StackTop)
                    StackTop = StackTop - 1 'Pop.
                Else
                    Err.Raise &H80044740, , "Bad expression"
                End If
            Else
                On Token - (PARENS_LAST + 1) _
                    GoTo Subtract, Multiply, Divide, IntDivide, Modulus, Power
'Add:
                Token = Stack(StackTop - 1) + Stack(StackTop)
                GoTo PopPop
Subtract:
                Token = Stack(StackTop - 1) - Stack(StackTop)
                GoTo PopPop
Multiply:
                Token = Stack(StackTop - 1) * Stack(StackTop)
                GoTo PopPop
Divide:
                Token = Stack(StackTop - 1) / Stack(StackTop)
                GoTo PopPop
IntDivide:
                Token = Stack(StackTop - 1) \ Stack(StackTop)
                GoTo PopPop
Modulus:
                Token = Stack(StackTop - 1) Mod Stack(StackTop)
                GoTo PopPop
Power:
                Token = Stack(StackTop - 1) ^ Stack(StackTop)
                'GoTo PopPop
PopPop:
                StackTop = StackTop - 2 'Pop, pop.
            End If
        End If
        'Push Token onto the stack:
        StackTop = StackTop + 1
        If StackTop > UBound(Stack) Then
            ReDim Preserve Stack(UBound(Stack) + STK_GROWBY)
        End If
        Stack(StackTop) = Token
    Next
    If StackTop <> 0 Then Err.Raise &H80044760, , "Bad expression"
    EvaluateOutput = Stack(0)
End Function

Private Sub ParseToPostfix()
    'Uses Edsger W. Dijkstra's "Shunting-yard algorithm."
    Dim Token As Variant
    Dim TokenType As VbVarType
    Dim FoundMatchingParenLeft As Boolean
    Dim OpPrec As Byte

    'Clear queue and stack:
    QueueLast = -1
    StackTop = -1

    Token = ParseToken()
    TokenType = VarType(Token)
    Do Until TokenType = vbEmpty
        If TokenType = vbByte Then
            'Operator or parenthesis:
            On Token + 1 GoTo ParenLeft, ParenRight
'Operator:
            If StackTop >= 0 Then
                OpPrec = OperatorPrecedences(Token)
                Do While StackTop >= 0
                    If OperatorPrecedences(Stack(StackTop)) >= OpPrec Then
                        If Stack(StackTop) = PAREN_LEFT Then
                            Exit Do
                        Else
                            'Queue stack-top item:
                            QueueLast = QueueLast + 1
                            If QueueLast > UBound(Queue) Then
                                ReDim Preserve Queue(UBound(Queue) + Q_GROWBY)
                            End If
                            Queue(QueueLast) = Stack(StackTop)
                            StackTop = StackTop - 1 'Pop.
                        End If
                    Else
                        Exit Do
                    End If
                Loop
            End If
            'Push Token onto the stack:
            StackTop = StackTop + 1
            If StackTop > UBound(Stack) Then
                ReDim Preserve Stack(UBound(Stack) + Q_GROWBY)
            End If
            Stack(StackTop) = Token
            GoTo NextToken
ParenLeft:
            'Push Token onto the stack:
            StackTop = StackTop + 1
            If StackTop > UBound(Stack) Then
                ReDim Preserve Stack(UBound(Stack) + STK_GROWBY)
            End If
            Stack(StackTop) = Token
            GoTo NextToken
ParenRight:
            FoundMatchingParenLeft = False
            Do While StackTop >= 0
                If Stack(StackTop) = PAREN_LEFT Then
                    StackTop = StackTop - 1 'Pop.
                    FoundMatchingParenLeft = True
                    Exit Do
                Else
                    'Queue stack-top item:
                    QueueLast = QueueLast + 1
                    If QueueLast > UBound(Queue) Then
                        ReDim Preserve Queue(UBound(Queue) + Q_GROWBY)
                    End If
                    Queue(QueueLast) = Stack(StackTop)
                    StackTop = StackTop - 1 'Pop.
                End If
            Loop
            If Not FoundMatchingParenLeft Then
                Err.Raise &H80044780, , "Mismatched parentheses"
            End If
            'GoTo NextToken
        Else
            'Numeric value, queue Token:
            QueueLast = QueueLast + 1
            If QueueLast > UBound(Queue) Then
                ReDim Preserve Queue(UBound(Queue) + Q_GROWBY)
            End If
            Queue(QueueLast) = Token
        End If
NextToken:
        Token = ParseToken()
        TokenType = VarType(Token)
    Loop
    Do While StackTop >= 0
        If Stack(StackTop) <= PARENS_LAST Then
            'Unexpected parenthesis:
            Err.Raise &H80044780, , "Mismatched parentheses"
        Else
            'Queue stack-top item:
            QueueLast = QueueLast + 1
            If QueueLast > UBound(Queue) Then
                ReDim Preserve Queue(UBound(Queue) + Q_GROWBY)
            End If
            Queue(QueueLast) = Stack(StackTop)
            StackTop = StackTop - 1 'Pop.
        End If
    Loop
End Sub

Private Function ParseToken() As Variant
    Dim CharsUsed As Long
    Dim HRESULT As HRESULT
    Dim Operator As Integer
    Dim DecodedOp As Byte

    CharsUsed = StrSpn(pNextChar, pWhiteSpace)
    pNextChar = pNextChar + 2 * CharsUsed
    CharsLeft = CharsLeft - CharsUsed
    If CharsLeft = 0 Then
        Exit Function 'Return Empty.
    End If

    If OperandExpected Then
        NUMPARSE.cDig = MAX_DIGITS
        HRESULT = VarParseNumFromStr(pNextChar, LCID, 0, NUMPARSE, rgbDig(0))
    Else
        GoTo ParseOperator
    End If
    Select Case HRESULT
        Case S_OK
            HRESULT = VarNumFromParseNum(NUMPARSE, rgbDig(0), ParsedTypes, ParseToken)
            Select Case HRESULT
                Case S_OK
                    CharsUsed = NUMPARSE.cchUsed
                    pNextChar = pNextChar + 2 * CharsUsed
                    CharsLeft = CharsLeft - CharsUsed
                    OperandExpected = False
                    Exit Function
                Case E_OUTOFMEMORY
                    Err.Raise 7 'Out of memory: rgbDig() too small, i.e. too many digits.
                Case DISP_E_TYPEMISMATCH
                    Err.Raise 13 'Type mismatch.
                Case DISP_E_OVERFLOW
                    Err.Raise 6 'Overflow.
                Case Else
                    Err.Raise &H800447A0, _
                              , _
                              "Error &H" & Right$(String$(7, "0") & Hex$(HRESULT), 8)
            End Select
        Case DISP_E_TYPEMISMATCH
            'Didn't parse, may be an operator.
ParseOperator:
            GetMem2 pNextChar, Operator
            DecodedOp = OperatorVals(Operator)
            If DecodedOp <> INVALID_OPERATOR Then
                ParseToken = DecodedOp
                pNextChar = pNextChar + 2
                CharsLeft = CharsLeft - 1
                OperandExpected = DecodedOp <> PAREN_RIGHT
            Else
                Err.Raise &H80044790, , "Bad expression"
            End If
        Case E_OUTOFMEMORY
            Err.Raise 7 'Out of memory: rgbDig() too small, i.e. too many digits.
        Case DISP_E_OVERFLOW
            Err.Raise 6 'Overflow.
        Case Else
            Err.Raise &H80044700, , "Error &H" & Right$(String$(7, "0") & Hex$(HRESULT), 8)
    End Select
End Function

Private Sub Class_Initialize()
    Dim I As Long
    Dim Ops() As Byte

    ReDim Queue(Q_INITIAL_SIZE - 1)
    ReDim Stack(STK_INITIAL_SIZE - 1)
    ReDim rgbDig(MAX_DIGITS - 1)
    NUMPARSE.dwInFlags = PARSE_FLAGS_DEFAULT
    ParsedTypes = [_vtbValid]
    pWhiteSpace = StrPtr(WHITE_SPACE)
    'Build the OperatorVals() lookup table:
    For I = 0 To UBound(OperatorVals)
        OperatorVals(I) = INVALID_OPERATOR
    Next
    Ops = StrConv(OPERATORS, vbFromUnicode)
    For I = 0 To UBound(Ops)
        OperatorVals(Ops(I)) = I
    Next
    OperatorPrecedences = StrConv(OPERATOR_PRECEDENCES, vbFromUnicode)
End Sub
