VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "VFormat"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = True
Attribute VB_Exposed = False
Option Explicit
'=======
'VFormat version 1.1
'=======
'
'A VB6 class used as a VBA.Format$() alternative using a specific locale
'instead of the current process locale.
'
'This is a VB_PredeclaredId = True class, so you do not need to create
'private instances unless you want different settings from those of the
'global predeclared instance.
'
'This is most useful with LCID_INVARIANT for numeric data that is
'perisisted or communicated as String data.  This avoids cross-cultural
'factors related to language, the decimal point character, grouping
' character, etc.
    
Private Const S_OK As Long = 0&
Private Const DISP_E_BUFFERTOOSMALL As Long = &H80020013
Private Const DISP_E_TYPEMISMATCH As Long = &H80020005
Private Const E_INVALIDARG As Long = &H80000003
Private Const E_OUTOFMEMORY As Long = &H80000002

Private Declare Function VarTokenizeFormatString Lib "oleaut32" ( _
    ByVal pstrFormat As Long, _
    ByRef rgbTok As Byte, _
    ByVal cbTok As Long, _
    ByVal iFirstDay As VbDayOfWeek, _
    ByVal iFirstWeek As VbFirstWeekOfYear, _
    ByVal LCID As Long, _
    Optional ByVal pcbActual As Long) As Long

Private Declare Function VarFormatFromTokens Lib "oleaut32" ( _
    ByRef pvarIn As Variant, _
    ByVal pstrFormat As Long, _
    ByRef bTokCur As Byte, _
    ByVal dwFlags As Long, _
    ByVal pbstrOut As Long, _
    ByVal LCID As Long) As Long
    
Private Const TOK_BUF_CHUNK As Long = 60

Private TokBufSize As Long
Private TokBuf() As Byte

Private Const SORT_DEFAULT As Long = &H0& 'sorting default

Private Const LANG_NEUTRAL As Long = &H0&
Private Const LANG_INVARIANT As Long = &H7F&

Private Const SUBLANG_NEUTRAL As Long = &H0&  'language neutral
Private Const SUBLANG_DEFAULT As Long = &H1&  'user default
Private Const SUBLANG_SYS_DEFAULT As Long = &H2&  'system default

Private Const LANG_SYSTEM_DEFAULT As Long = LANG_NEUTRAL Or SUBLANG_SYS_DEFAULT * &H400&
Private Const LANG_USER_DEFAULT As Long = LANG_NEUTRAL Or SUBLANG_DEFAULT * &H400&

Public Enum LCIDs
    LCID_SYSTEM_DEFAULT = LANG_SYSTEM_DEFAULT Or SORT_DEFAULT * &H10000
    LCID_USER_DEFAULT = LANG_USER_DEFAULT Or SORT_DEFAULT * &H10000
    LCID_NEUTRAL = (LANG_NEUTRAL Or SUBLANG_NEUTRAL * &H400&) Or SORT_DEFAULT * &H10000
    LCID_INVARIANT = (LANG_INVARIANT Or SUBLANG_NEUTRAL * &H400&) Or SORT_DEFAULT * &H10000

    'Add more as you need them here:
    LCID_ENUK = &H809&
    LCID_ENUS = &H409&
    LCID_DEDE = &H407& 'German, Germany.
    LCID_FIFI = &H40B& 'Finland, Finnish.
    LCID_FRFR = &H40C& 'France, French.
End Enum
#If False Then
Dim LCID_SYSTEM_DEFAULT, LCID_USER_DEFAULT, LCID_NEUTRAL, LCID_INVARIANT
Dim LCID_ENUK, LCID_ENUS, LCID_DEDE, LCID_FIFI, LCID_FRFR
#End If

Public LCID As LCIDs

'Default member:
Public Function Format( _
    ByVal Value As Variant, _
    Optional ByVal FormatString As String) As String
Attribute Format.VB_UserMemId = 0

    Dim Result As Long

    Do
        Result = VarTokenizeFormatString(StrPtr(FormatString), _
                                         TokBuf(0), _
                                         TokBufSize, _
                                         vbSunday, _
                                         vbFirstJan1, _
                                         LCID)
        Select Case Result
            Case S_OK
                'Fall through.
            Case DISP_E_BUFFERTOOSMALL
                TokBufSize = TokBufSize + TOK_BUF_CHUNK
                ReDim TokBuf(TokBufSize - 1)
            Case Else
                'Error.
                Err.Raise 5, TypeName(Me) 'Invalid procedure call or argument.
                Exit Function
        End Select
    Loop Until Result = S_OK

    If VarFormatFromTokens(Value, _
                           StrPtr(FormatString), _
                           TokBuf(0), _
                           0, _
                           VarPtr(Format), _
                           LCID) Then Err.Raise 5 'Invalid procedure call or argument.
End Function

Private Sub Class_Initialize()
    TokBufSize = TOK_BUF_CHUNK
    ReDim TokBuf(TokBufSize - 1)
    LCID = LCID_INVARIANT
End Sub
